INDETERMINABLE_TEST_CASE = 0
FAILED_TEST_CASE = 1
SUCCESSFUL_TEST_CASE = 2

NUM_LABELS = 3

DATA_DIR = "data/"
EXECUTION_TIME_DIR = "execution_time/"
CASE_DIR = "case/"

AUDIO_DIR = "audio/"
TRANSCRIPTION_DIR = "transcription/"
